'=========================================================================
' Prozeduren fuer Diagramme (Graphik mit 8x16 Zeichensatz)
' GRAPHIK.BAS
' Uebernommen aus QuickBASIC Januar 2006
' Letzte Ueberarbeitung 25.01.12
'=========================================================================


#INCLUDE ONCE "COMMON.bi"


DECLARE SUB OutText(x AS DOUBLE, y AS DOUBLE, Text AS STRING, horPos AS STRING, _
  vertPos AS STRING)
DECLARE SUB Koordinaten(xmin AS DOUBLE, ymin AS DOUBLE, xmax AS DOUBLE, ymax AS DOUBLE, _
  nx AS INTEGER, ny AS INTEGER, hor AS STRING, vert AS STRING, Text AS STRING)                    
DECLARE SUB ZeichneKurve(x() AS DOUBLE, y() AS DOUBLE, Farbe AS INTEGER)
DECLARE SUB ZeichneFkt(fkt AS FUNCTION (x As DOUBLE) AS DOUBLE, xmin AS DOUBLE, _
  xmax AS DOUBLE, Farbe AS INTEGER)
DECLARE SUB Balkendiagramm(x() AS STRING, y() AS DOUBLE, ymin AS DOUBLE, ymax AS DOUBLE, _
  ny AS INTEGER, hor AS STRING, vert AS STRING, Text AS STRING, Farbe AS INTEGER)
DECLARE SUB Kreisdiagramm(x() AS STRING, y() AS DOUBLE, Farbe() AS INTEGER, Text AS STRING)
DECLARE FUNCTION RGBWert(x AS DOUBLE) AS UINTEGER
DECLARE SUB PlotFkt(fkt AS FUNCTION (x As DOUBLE, y As DOUBLE) AS DOUBLE, _
  xmin AS DOUBLE, ymin AS DOUBLE, xmax AS DOUBLE, ymax AS DOUBLE, Koord AS INTEGER = 0)

  
SUB OutText(x AS DOUBLE, y AS DOUBLE, Text AS STRING, horPos AS STRING, _
  vertPos AS STRING)
' Gibt einen Text an pixelgenauen Koordinaten aus. Parameter:
' - x, y = Koordinaten
' - Text = auszugebender Text
' - horPos = horizontale Position des Textes bezueglich x:
'   "r": Text rechts von x
'   "l": Text links von x
'   "z": Text zentriert zu x
' - vertPos = vertikale Position des Textes bezueglich y:
'   "u": Text unterhalb von y
'   "o": Text oberhalb von y
'   "z": Text zentriert zu y
' OutText(x, y, Text, "r", "u") funktioniert also genau so wie 
' DRAW STRING (y, y), Text.


  DIM AS DOUBLE xp, yp
  
  'Falls x und y Sichtfensterkoordinaten sind, werden diese in physiche
  'Koordinaten umgerechnet:
  xp = PMAP(x, 0)
  yp = PMAP(y, 1)
  
  'xp entsprechend der horizontalen Position verschieben:
  SELECT CASE LCASE(horPos)
  CASE "l": xp = xp - 8 * LEN(Text)
  CASE "z": xp = xp - 8 * LEN(Text) / 2
  END SELECT
  
  'yp entsprechend der vertikalen Position verschieben:
  SELECT CASE LCASE(vertPos)
  CASE "o": yp = yp - 16
  CASE "z": yp = yp - 8
  END SELECT
  
  'xp und yp wieder in Sichtfensterkordinaten umrechnen:
  xp = PMAP(xp, 2)
  yp = PMAP(yp, 3)

  'Text plazieren:
  DRAW STRING (xp, yp), Text
  
END SUB
                       

SUB Koordinaten(xmin AS DOUBLE, ymin AS DOUBLE, xmax AS DOUBLE, ymax AS DOUBLE, _
  nx AS INTEGER, ny AS INTEGER, hor AS STRING, vert AS STRING, Text AS STRING)
' Koordinatenfeld mit linearen Achsenteilungen.
' Parameter:
' - xmin ... ymax definieren die mathematischen Abmessungen des Koordinatenfeldes.
' - nx, ny = Zahl der Achsenabschnitte auf der x- bzw. y-Achse
' - hor, vert = Beschriftung fuer x-Achse bzw. y-Achse
' - Text = Bildunterschrift

  DIM AS INTEGER i, xlo, ylo, xru, yru
  DIM AS DOUBLE ax, ay, x, y
  
  'Physikalische Koordinaten der linken oberen und unteren rechten Ecke des Diagrammfeldes:
  xlo = 120
  ylo = 20
  xru = 560
  yru = 320
  
  VIEW
  WINDOW SCREEN
  
  'Beschriftung vertikale Achse:
  OutText(xlo, ylo, STR(CSNG(ymax)) + SPACE(1), "l", "z")
  OutText(xlo, yru, STR(CSNG(ymin)) + SPACE(1), "l", "z")
  OutText(xlo, (ylo + yru) / 2, vert + SPACE(1), "l", "z")  
  
  'Beschriftung horizontale Achse:
  OutText(xlo, yru + 8, STR(CSNG(xmin)), "z", "u")
  OutText(xru, yru + 8, STR(CSNG(xmax)), "z", "u")
  OutText((xlo + xru) / 2 , yru + 8, hor, "z", "u")
  
  'Bildunterschrift:
  OutText((xlo + xru) / 2 , yru + 24, Text, "z", "u")
  
  'Fenster fuer Koordinatensystem in Sichtkoordinaten definieren:
  VIEW (xlo, ylo) - (xru, yru),, LOWORD(COLOR)
  WINDOW (xmin, ymin)-(xmax, ymax)
    
  'Teilung horizontale Achse:
  ax = (xmax - xmin) / nx
  FOR i = 0 TO nx
    x = xmin + i * ax: LINE (x, ymin)-(x, ymax)
  NEXT i

  'Teilung vertikale Achse:
  ay = (ymax - ymin) / ny
  FOR i = 0 TO ny
    y = ymin + i * ay: LINE (xmin, y)-(xmax, y)
  NEXT i

END SUB


SUB ZeichneKurve (x() AS DOUBLE, y() AS DOUBLE, Farbe AS INTEGER)
' Kurve durch die Punkte x(0), y(0) bis x(n), y(n) als Polygonzug in der
' angegebenen Farbe, z.B. in das mit der SUB Koordinaten gezeichnete
' Koordinatensystem.

  DIM AS INTEGER n, i
  
  n = UBOUND(x)
  FOR i = 1 TO n
    LINE (x(i - 1), y(i - 1))-(x(i), y(i)), Farbe
  NEXT i

END SUB


SUB ZeichneFkt(fkt AS FUNCTION (x As DOUBLE) AS DOUBLE, xmin AS DOUBLE, _
  xmax AS DOUBLE, Farbe AS INTEGER)
' Zeichnet ein Funktion fkt(x) im Wertebereich xmin bis xmax in der angegebenen
' Farbe, z.B. in das mit der SUB Koordinaten gezeichnete Koordinatensystem.

  DIM AS DOUBLE dx, x
  
  dx = (xmax - xmin) / 100
  FOR x = xmin TO xmax - dx STEP dx
    LINE (x, fkt(x)) - (x + dx, fkt(x + dx)), Farbe
  NEXT x
  
END SUB


SUB Balkendiagramm(x() AS STRING, y() AS DOUBLE, ymin AS DOUBLE, ymax AS DOUBLE,_
  ny AS INTEGER, hor AS STRING, vert AS STRING, Text AS STRING, Farbe AS INTEGER)
' - x(0) bis x(n) = Benennung der Balken
' - y(0) bis y(n) = Wert der Balken
' - n + 1         = Anzahl der Balken
' - ymin, ymax    = Abmessung vertikale Achse (ymin muss <= 0 sein!)
' - ny            = Anzahl der Achsenabschnitte auf der vertikalen Achse
' - hor           = Beschriftung horizontale Achse
' - vert          = Beschriftung vertikale Achse (max 13 Zeichen)
' - Text          = Bildunterschrift
' - Farbe         = Farbe

  DIM AS INTEGER n = UBOUND(x), i, xlo, ylo, xru, yru
  DIM AS DOUBLE xmin, xmax, ay, y0, Breite
  
  'Physikalische Koordinaten der linken oberen und unteren rechten Ecke des Diagrammfeldes:
  xlo = 120
  ylo = 20
  xru = 560
  yru = 320

  VIEW
  WINDOW SCREEN
  
  'Beschriftung vertikale Achse:
  OutText(xlo, ylo, STR(CSNG(ymax)) + SPACE(1), "l", "z")
  OutText(xlo, yru, STR(CSNG(ymin)) + SPACE(1), "l", "z")
  OutText(xlo, (ylo + yru) / 2, vert + SPACE(1), "l", "z")
  
  'Beschriftung horizontale Achse:
  FOR i = 0 TO n
    OutText(xlo + (xru - xlo) * (i + 0.5) / (n + 1), yru + 8, x(i), "z", "u")
  NEXT i
  OutText((xlo + xru) / 2, yru + 24, hor, "z", "u")
    
  'Bildunterschrift:
  OutText((xlo + xru) / 2, yru + 40, Text, "z", "u")
  
  'Fenster fuer Graphik in Sichtkoordinaten definieren:
  VIEW (xlo, ylo)-(xru, yru),,LOWORD(COLOR)
  xmin = -0.5: xmax = n + 0.5
  WINDOW (xmin, ymin)-(xmax, ymax)

  'Teilung vertikale Achse:
  ay = (ymax - ymin) / ny
  FOR i = 0 TO ny
    y0 = ymin + i * ay
    LINE (xmin, y0)-(xmax, y0)
  NEXT i

  'Balken zeichnen:
  FOR i = 0 TO n
    LINE(i - 0.4, 0) - (i + 0.4, y(i)), Farbe, BF
  NEXT i

END SUB


SUB Kreisdiagramm(x() AS STRING, y() AS DOUBLE, Farbe() AS INTEGER, Text AS STRING)
' - x(0) bis x(n) = Bezeichnung der Kreissegmente
' - y(0) bis y(n) = Wert der Kreissegmente
' - Farbe(0) bis Farbe(n) = Farben der Kreissegmente 
' - n + 1         = Anzahl der Kreissegmente
' - Text          = Bildunterschrift

  DIM AS INTEGER n = UBOUND(x), i, x0, y0, r
  DIM AS DOUBLE alpha(n), sum, xp, yp, beta
  
  'Kreismittelpunkt und -radius:
  x0 = 320
  y0 = 200
  r = 140
  
  VIEW
  WINDOW SCREEN
  
  'Diagramm berechnen und segmentweise zeichnen (alpha(0) bis alpha(n) sind
  'die zu y(0) bis y(n) zugehoerigen Winkel im Kreis:
  sum = 0
  FOR i = 0 TO n
    sum = sum + y(i)
  NEXT i
  'Die Winkel alpha(i) zaehlen ab pi/2 (oben) im Uhrzeigersinn, also entgegen-
  'gesetzt zum mathematischen Drehsinn:
  alpha(0) = pi / 2 - 2 * pi * y(0) / sum
  FOR i = 1 TO n
    alpha(i) = alpha(i - 1) - 2 * pi * y(i) / sum
  NEXT i
  CIRCLE (x0, y0), r
  LINE (x0, y0) - (x0, y0 - r)
  FOR i = 0 TO n
    LINE (x0, y0) - (x0 + r * COS(alpha(i)), y0 - r * SIN(alpha(i)))
  NEXT i
    
  'Diagramm-Segmente faerben und beschriften:
  FOR i = 0 TO n
    IF i = 0 THEN beta = (pi / 2 + alpha(0)) / 2 _
      ELSE beta = (alpha(i) + alpha(i - 1)) / 2
    xp = x0 + 0.5 * r * COS(beta)
    yp = y0 - 0.5 * r * SIN(beta)
    PAINT (xp, yp), Farbe(i), LOWORD(COLOR)    
    xp = x0 + 1.1 * r * COS(beta)
    yp = y0 - 1.1 * r * SIN(beta)
    IF beta < -pi / 2 THEN
      OutText(xp, yp, x(i) + ": " + STR(CSNG(y(i))), "l", "z")
    ELSE
      OutText(xp, yp, x(i) + ": " + STR(CSNG(y(i))), "r", "z")
    END IF
  NEXT i
  
  'Bildunterschrift:
  OutText(x0, y0 + 1.3 * r, Text, "z", "o")
  
END SUB
                   
                          
FUNCTION RGBWert(z AS DOUBLE) AS UINTEGER
' Berechnet eine RGB-Farbnummer abhaengig vom Wert z
' (Wertebereich fuer k = 0 bis 10, Farbskala blau - blaugruen - gruen - gelb - rot)

  IF z < 0 THEN
    RETURN RGB(0, 0, 250)
  ELSEIF z < 2.5 THEN
    RETURN RGB(0, 100 * z, 250)
  ELSEIF z < 5 THEN 
    RETURN RGB(0, 250, 250 - 100 * (z - 2.5))
  ELSEIF z < 7.5 THEN
    RETURN RGB(100 * (z - 5), 250, 0)
  ELSEIF z < 10 THEN
    RETURN RGB(250, 250 - 100 * (z - 7.5), 0)
  ELSE
    RETURN RGB(250, 0, 0)
  END IF
  
END FUNCTION


SUB PlotFkt(fkt AS FUNCTION (x As DOUBLE, y As DOUBLE) AS DOUBLE, _
  xmin AS DOUBLE, ymin AS DOUBLE, xmax AS DOUBLE, ymax AS DOUBLE, Koord AS INTEGER = 0)
' Zeichnet eine Funktion fkt(x, y) als "Hoehenkarte" (Graphik mit
' 32 Bit Farbtiefe!) in 20 Stufen.
' Parameter:
' - xmin, xmax, ymin, ymax = Wertebereich von x und y
' Optionaler Parameter:
' - Koord = 0: ganzes Graphikfenster als Diagrammflaeche
'   Koord <> 0: Diagrammflaeche wie bei SUB Koordinaten in GRAPHIK.bas. Mit der 
'   SUB Koordinaten kann dann ein Koordinatensystem ueber das Diagramm gelegt werden.
' Benoetigte Prozedur: FUNCTION fkt (x, y) (im Hauptmodul zu definieren)
' Aufruf der SUB im Hauptmodul: PlotFkt(@fkt, (sonstige Parameter))

  DIM AS INTEGER xlo, ylo, xru, yru
  DIM AS DOUBLE x, y, dx, dy, z, zmin, zmax
  
  'Physikalische Koordinaten der linken oberen und unteren rechten Ecke des Diagrammfeldes:
  IF Koord <> 0 THEN
  	xlo = 120
  	ylo = 20
  	xru = 560
  	yru = 320
  ELSE
  	xlo = 1
  	ylo = 1
  	xru = 640
  	yru = 480
  EndIf
  
  View
  WINDOW SCREEN
  
  VIEW (xlo, ylo) - (xru, yru),, LOWORD(COLOR)
  WINDOW (xmin, ymin) - (xmax, ymax)

  'Pixelgenaue Intervalle fuer x und y:
  dx = (xmax - xmin) / (xru - xlo + 1)
  dy = (ymax - ymin) / (yru - ylo + 1)
  
  'Minimalen und maximalen Wert von z suchen:
  zmin = 0
  zmax = 0
  FOR x = xmin TO xmax
    FOR y = ymin TO ymax
      z = fkt(x, y)
      IF z < zmin THEN zmin = z
      IF z > zmax THEN zmax = z
    NEXT y
  NEXT x
  
  'Zeichnen:
  FOR x = xmin TO xmax STEP dx
    FOR y = ymax TO ymin STEP -dy
      z = fkt(x, y)
      PSET (x, y), RGBWert(0.5 * CINT((z - zmin) * 20 / (zmax - zmin)))
    NEXT y
  NEXT x  
    
END SUB
